"""@package ecohydrolib.ssurgo.attributequery
    
@brief Make tabular queries against USDA Soil Data Mart SOAP web service interface
@note Requires python-httplib2 to be installed, else requests to soil data mart may timeout

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import cStringIO
import xml.sax
import json

import numpy as np
import socket
import httplib2
from oset import oset

from saxhandlers import SSURGOMUKEYQueryHandler

_BUFF_LEN = 4096 * 10

ATTRIBUTE_NAMESPACE = 'ms'
ATTRIBUTE_LIST = ['ksat', 'pctClay', 'pctSilt', 'pctSand', 'porosity',
                 'pmgroupname', 'texture', 'tecdesc', 'fieldCap', 
                 'avlWatCap']
ATTRIBUTE_LIST_NUMERIC = ['ksat', 'pctClay', 'pctSilt', 'pctSand', 'porosity', 
                        'fieldCap', 'avlWatCap']
# DERIVED_ATTRIBUTES must be a dictionary whose values are valid Python expressions combining members of ATTRIBUTE_LIST_NUMERIC
DERIVED_ATTRIBUTES = { 'drnWatCont': 'porosity - fieldCap' }
ATTRIBUTE_LIST_NUMERIC.extend( DERIVED_ATTRIBUTES.keys() )

def strListToString(strList):
    """ Converts a Python list of string values into a string containing quoted, 
        comma separated representation of the list.
        
        @param strList List of strings
        
        @return String containing quoted, comma separated representation of the list
    """
    numStr = len(strList)
    assert(numStr > 0)
    
    output = cStringIO.StringIO()
    for i in range(numStr - 1):
        output.write("'%s'," % (strList[i],))
    output.write("'%s'" % (strList[numStr-1],))
    
    returnStr = output.getvalue()
    output.close()
    return returnStr


def computeWeightedAverageKsatClaySandSilt(soilAttrTuple):
    """ Computes weighted average for Ksat, %clay/silt/sand for a SSURGO mukey based on values
        for each component in the mukey; weights based on component.comppct_r.
    
        @param soilAttrTuple Tuple returned from getParentMatKsatTexturePercentClaySiltSandForComponentsInMUKEYs
    
        @return Tuple containing: (1) a list containing column names; (2) a list of lists containing averaged soil properties for each mukey
    """
    data = list()
    representativeComponentDict = dict()
    derivedSet = oset()
    idx = 0
    
    # Convert numbers as text to numbers
    for row in soilAttrTuple[1]:
        mukey = int(row[0])
        comppct_r = int(row[2])
        try:
            maxRepComp = representativeComponentDict[mukey][1]
            if maxRepComp < comppct_r:
                representativeComponentDict[mukey] = (idx, comppct_r)
        except KeyError:
            representativeComponentDict[mukey] = (idx, comppct_r)
        try:
            hzdept_r = float(row[7])
        except ValueError:
            hzdept_r = -1
        try:
            ksat_r = float(row[8])
        except ValueError:
            ksat_r = -1
        try:
            claytotal_r = float(row[9])
        except ValueError:
            claytotal_r = -1
        try:
            silttotal_r = float(row[10])
        except ValueError:
            silttotal_r = -1
        try:
            sandtotal_r = float(row[11])
        except ValueError:
            sandtotal_r = -1
        try:
            wsatiated_r = float(row[12]) # a.k.a. porosity
        except ValueError:
            wsatiated_r = -1
        try:
            wthirdbar_r = float(row[13]) # a.k.a. field capacity
        except ValueError:
            wthirdbar_r = -1
        try:
            awc_r = float(row[14]) # a.k.a. plant available water capacity
        except ValueError:
            awc_r = -1
    
        data.append([mukey, row[1], comppct_r, row[3], row[4], row[5], row[6], hzdept_r, 
                     ksat_r, claytotal_r, silttotal_r, sandtotal_r, wsatiated_r,
                     wthirdbar_r, awc_r])
        idx = idx + 1

    mukeyCol = [row[0] for row in data]
    comppctCol = [row[2] for row in data]
    ksatCol = [row[8] for row in data]
    clayCol = [row[9] for row in data]
    siltCol = [row[10] for row in data]
    sandCol = [row[11] for row in data]
    porosityCol = [row[12] for row in data]
    fieldCapCol = [row[13] for row in data]
    availWaterCapCol = [row[14] for row in data]

    # Put values into Numpy 2-D array    
    npdata = np.array([mukeyCol, comppctCol, ksatCol, clayCol, siltCol, sandCol, 
                       porosityCol, fieldCapCol, availWaterCapCol]).transpose()
    # Remove duplicate rows 
    #   (which will arise because there can be multiple parent material groups for a given component)
    npdata = np.array([np.array(x) for x in set(tuple(x) for x in npdata)])
    # Register NoData values
    npdata = np.ma.masked_where(npdata == -1, npdata)

    # Calculate weighted average using component.comppct_r as weights
    avgSoilAttr = list()
    mukeySet = set(mukeyCol)
    for mukey in mukeySet:
        mySubSet = npdata[npdata[:,0] == mukey]
        myComppct = mySubSet[:,1]
        myKsat = mySubSet[:,2]
        myClay = mySubSet[:,3]
        mySilt = mySubSet[:,4]
        mySand = mySubSet[:,5]
        myPorosity = mySubSet[:,6]
        myFieldCap = mySubSet[:,7]
        myAvailWaterCap = mySubSet[:,8]
        # Calculate weighted averages, ignoring NoData values
        # These variable names MUST match values in ATTRIBUTE_LIST_NUMERIC
        ksat = np.ma.average(myKsat, weights=myComppct)
        pctClay = np.ma.average(myClay, weights=myComppct)
        pctSilt = np.ma.average(mySilt, weights=myComppct)
        pctSand = np.ma.average(mySand, weights=myComppct)
        porosity = np.ma.average(myPorosity, weights=myComppct)
        fieldCap = np.ma.average(myFieldCap, weights=myComppct)
        avlWatCap = np.ma.average(myAvailWaterCap, weights=myComppct)
        
        # Get modal value for qualitative values (pmgroupname, texture, tecdesc)
        maxRepIdx = representativeComponentDict[mukey][0]
        pmgroupname = data[maxRepIdx][3]
        texture = data[maxRepIdx][4]
        texdesc = data[maxRepIdx][5]
        
        attrList = [mukey, ksat, pctClay, pctSilt, pctSand, porosity, pmgroupname, texture, texdesc,
                    fieldCap, avlWatCap]
        # Generate derived variables
        for attr in DERIVED_ATTRIBUTES.keys():
            derivedAttr = eval( DERIVED_ATTRIBUTES[attr] )
            derivedSet.add(attr)
            attrList.append(derivedAttr) 
        
        avgSoilAttr.append(attrList)
    avgSoilHeaders = list(ATTRIBUTE_LIST)
    avgSoilHeaders.insert(0, 'mukey')
    for derived in derivedSet:
        print("Computed derived attribute %s = %s" % \
              (derived, DERIVED_ATTRIBUTES[derived]) )
        avgSoilHeaders.append(derived)
    
    return (avgSoilHeaders, avgSoilAttr)


def joinSSURGOAttributesToFeaturesByMUKEY_GeoJSON(geojson, typeName, ssurgoAttributes):
    """ Join SSURGO tabular attributes to MapunitPoly or MapunitPolyExtended features based on
        MUKEY.
    
        @param geojson JSON object representing SSURGO MapunitPolyExtended or MapunitPoly features
        @param typeName String of either 'MapunitPoly' or 'MapunitPolyExtended'
        @param ssurgoAttributes Tuple containing two lists: (1) list of column names; (2) list of
        column values.  Assumes the following column names and order:
        ['mukey', 'avgKsat', 'avgClay', 'avgSilt', 'avgSand', 'avgPorosity']
    
        @note geojson JSON object will be modified by this function   
    """
    assert(typeName == 'MapunitPoly' or typeName == 'MapunitPolyExtended')
    
    # Index attributes by MUKEY
    attributeDict = {}
    idx = 0
    for row in ssurgoAttributes[1]:
        myMukey = row[0]
        attributeDict[myMukey] = idx
        idx = idx + 1
        
    for feature in geojson['features']:
        properties = feature['properties']
        mukey = int(properties['mukey'])
        
        try:
            mukeyIdx = attributeDict[mukey]
        except KeyError:
            continue
        
        currAttrIdx = 1
        # Add attributes to this feature's properties table
        for attr in ssurgoAttributes[0][1:]:
            properties[attr] = str(ssurgoAttributes[1][mukeyIdx][currAttrIdx])
            currAttrIdx += 1
            

def getParentMatKsatTexturePercentClaySiltSandForComponentsInMUKEYs(mukeyList):
    """ Query USDA soil datamart tabular service for ksat, texture, % clay, % silt, % sand for all
        components in the specified map units.
    
        @param mukeyList List of strings representing the MUKEY of each map unit for which we would 
        like to query attributes.
    
        @return Tuple containing an ordered set (oset.oset) representing column names, and a list, 
        each element containing a list of column values for each row in the SSURGO query result for each map unit
        
        @raise socket.error if there was an error reading the data from the web service
        @raise Exception if webservice returned code other than 200
    """ 

    #client = SoapClient(wsdl="http://sdmdataaccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx?WSDL")

    # Query to get component %, ksat_r, texture, texture description, parent material, horizon name, horizon depth,
    # %clay, %silt, %sand, and porosity (as wsatiated.r, volumetric SWC at or near 0 bar tension) for all components in an MUKEY
    # Will select first non-organic horizon (i.e. horizon names that do not start with O, L, or F)
    QUERY_PROTO = """SELECT c.mukey, c.cokey, c.comppct_r, p.pmgroupname, tg.texture, tg.texdesc, 
ch.hzname, ch.hzdept_r, ch.ksat_r, ch.claytotal_r, ch.silttotal_r, ch.sandtotal_r, ch.wsatiated_r,
ch.wthirdbar_r, ch.awc_r
FROM component c
LEFT JOIN copmgrp p ON c.cokey=p.cokey AND p.rvindicator='yes'
INNER JOIN chorizon ch ON c.cokey=ch.cokey 
AND ch.hzdept_r=(SELECT TOP(1) hzdept_r FROM chorizon WHERE cokey=c.cokey AND (hzname NOT LIKE 'O' + '%%') and (hzname NOT LIKE 'L' + '%%') and (hzname NOT LIKE 'F' + '%%') ORDER BY hzdept_r ASC)
LEFT JOIN chtexturegrp tg ON ch.chkey=tg.chkey AND tg.rvindicator='yes' AND tg.texture<>'variable' AND tg.texture<>'VAR'
WHERE c.mukey IN (%s) ORDER BY c.cokey"""
    
    mukeyStr = strListToString(mukeyList)
    
    query = QUERY_PROTO % mukeyStr
    
    # Manually make SOAP query (it's a long story)
    host = 'sdmdataaccess.nrcs.usda.gov'
    url = '/Tabular/SDMTabularService.asmx'
    url = "http://" + host + url
    
    soapQueryBegin = """<?xml version="1.0" encoding="UTF-8"?><soap:Envelope xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"><soap:Header/><soap:Body>
    <RunQuery xmlns="http://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx">
    <Query>
    """
    soapQueryEnd = """</Query></RunQuery></soap:Body></soap:Envelope>"""
    
    soapQuery = soapQueryBegin + xml.sax.saxutils.escape(query) + soapQueryEnd
    
    headers = { 'SOAPAction': 'http://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx/RunQuery',  #'SOAPAction': 'RunQuery',
                'Content-Type': 'text/xml; charset=utf-8',
                'Content-length': str(len(soapQuery)) }
    h = httplib2.Http()
    res = None
    try:
        (res, content) = h.request(url, method='POST', body=soapQuery, headers=headers)
    except socket.error as e:
        raise e
    
    if 200 != res.status:
        raise Exception("Error %d encountered when reading SSURGO attributes from USDA webservice" % \
                        (res.status,) )

    # Parse results
    handler = SSURGOMUKEYQueryHandler()
    
    xml.sax.parseString(content, handler)

    return (handler.columnNames,handler.results)
    



"""@package ecohydrolib.hydro1k.basins
    
@brief Methods for querying HYDRO1k basins

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os
import sys
import errno


def getCatchmentShapefileForHYDRO1kBasins(config, outputDir, catchmentFilename, basins):
    """ Get shapefile (in WGS 84) for the drainage area associated with a list of HYDRO1k basins
    
        @note No return value. catchmentFilename will be written to outputDir if successful
        
        @param config A Python ConfigParser containing the following sections and options:
            'GDAL/OGR' and option 'PATH_OF_OGR2OGR' (absolute path of ogr2ogr binary)
            'HYDRO1k' and option 'PATH_OF_HYDRO1K_BAS' (absolute path to HYDRO1k North America na_bas.e00)
        @param outputDir String representing the absolute/relative path of the directory into which output 
            rasters should be written
        @param catchmentFilename String representing name of file to save catchment shapefile to
        @param basins List of strings representing basin IDs; must be of the same level: 1, 2, 3, 4, 5, 6
        where level corresponds to the number of characters in the ID string.
        
        @exception ConfigParser.NoSectionError
        @exception ConfigParser.NoOptionError
        @exception IOError(errno.ENOTDIR) if outputDir is not a directory
        @exception IOError(errno.EACCESS) if outputDir is not writable
        @exception Exception of catchment shapefile extraction fails
    """
    hydro1kPath = config.get('HYDRO1k', 'PATH_OF_HYDRO1K_BAS')
    if not os.access(hydro1kPath, os.R_OK):
        raise IOError(errno.EACCES, "The catchment feature layer at %s is not readable" %
                      hydro1kPath)
    hydro1kPath = os.path.abspath(hydro1kPath)
    
    hydro1kLayerName = config.get('HYDRO1k', 'HYDRO1k_BAS_LAYER_NAME')
    
    ogrCmdPath = config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
    if not os.access(ogrCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "The ogr2ogr binary at %s is not executable" %
                      ogrCmdPath)
    ogrCmdPath = os.path.abspath(ogrCmdPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    catchmentFilename = os.path.join(outputDir, catchmentFilename)
    
    assert(len(basins) > 0)
    # Ensure all basin IDs are of the same level
    levelNum = len(basins[0])
    assert( all( len(i) == levelNum for i in basins ) )
    levelStr = "LEVEL%d" % (levelNum,)
    
    # Build query for basin identifiers
    ogrCommand = "%s -t_srs EPSG:4326  -f 'ESRI Shapefile' -where '%s=%s" % \
        (ogrCmdPath, levelStr, basins[0])
    for basin in basins[1:]:
        ogrCommand = ogrCommand + " OR %s=%s" % (levelStr, basin) 
    ogrCommand = ogrCommand +"' " + catchmentFilename + " " + hydro1kPath + " " + hydro1kLayerName
      
    # Extract polygons for specified basins
    #sys.stderr.write("ogr command: %s\n" % ogrCommand)
    returnCode = os.system(ogrCommand)
    if returnCode != 0:
        raise Exception("OGR command %s failed." % (ogrCommand,))  


#!/usr/bin/env python
"""@package DumpClimateStationInfo

@brief Dump point climate station information from EcohydroLib metadata to standard output

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  

Pre conditions:
--------------
1. ClimatePointStation entries have been written to the climate point section of metadata associated with the project directory
   

Post conditions:
----------------
None

Usage:
@code
DumpClimateStationInfo.py -p /path/to/project_dir > file.csv
@endcode
"""
import sys
import os
import argparse

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata

parser = argparse.ArgumentParser(description='Dump point climate station information from EcohydroLib metadata to standard output')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory from which metadata should be read')
parser.add_argument('-s', '--separator', dest='separator', required=False, default=',', help='Field separator for output')
args = parser.parse_args()

context = Context(args.projectDir, None) 

s = args.separator

sys.stderr.write("Getting stations from metadata... ")
stations = GenericMetadata.readClimatePointStations(context)
sys.stderr.write("done\n")

for station in stations:
    output = station.id.upper() + s + str(station.latitude) + s + str(station.longitude) + s + str(station.elevation) + s + station.name + os.linesep
    sys.stdout.write(output)



#!/usr/bin/env python
"""@package RegisterRaster

@brief Register raster into metadata store for a project directory,
copying the raster into the project directory in the process.  Raster type must
be one of GenericMetadata.RASTER_TYPES.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'GDAL/OGR', 'PATH_OF_GDAL_TRANSLATE'
   'GDAL/OGR', 'PATH_OF_GDAL_WARP'

2. The following metadata entry(ies) must be present in the study area section of the metadata associated with the project directory:
   bbox_wgs84
   dem_res_x
   dem_res_y
   dem_columns
   dem_rows
   dem_srs

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   landcover [for GenericMetadata.RASTER_TYPE_LC; the name of the landcover raster] 
   roof_connectivity [for GenericMetadata.RASTER_TYPE_ROOF; the name of the roof raster]
   
2. Will write the following entry(ies) to the study area section of metadata associated with the project directory:
   landcover_type=custom [for GenericMetadata.RASTER_TYPE_LC]

Usage:
@code
RegisterRaster.py -p /path/to/project_dir -t RASTER_TYPE -r /raster/to/register
@endcode

@todo Set date in provenance to file modification date
"""
import os
import sys
import errno
import argparse
import textwrap

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.spatialdata.utils import bboxFromString
from ecohydrolib.spatialdata.utils import RASTER_RESAMPLE_METHOD
from ecohydrolib.spatialdata.utils import copyRasterToGeoTIFF
from ecohydrolib.spatialdata.utils import resampleRaster
from ecohydrolib.spatialdata.utils import getDimensionsForRaster
from ecohydrolib.spatialdata.utils import getSpatialReferenceForRaster

# Handle command line options
parser = argparse.ArgumentParser(description='Register raster dataset with project')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-t', '--type', dest='type', required=True, choices=GenericMetadata.RASTER_TYPES,
                    help='The type of raster dataset.')
parser.add_argument('-r', '--rasterfile', dest='rasterfile', required=True,
                    help='The relative/absolute path of the raster to be registered.')
parser.add_argument('-s', '--resampleMethod', dest='resampleMethod', required=False,
                    choices=RASTER_RESAMPLE_METHOD, default='near',
                    help="Method to use to resample raster to DEM extent and spatial reference (if necessary). Defaults to '%s'." % ('near',) )
parser.add_argument('-f', '--outfile', dest='outfile', required=False,
                    help='The name of the raster to be written to the project directory.  File extension ".tif" will be added.')
parser.add_argument('--force', dest='force', required=False, action='store_true',
                    help='Force registry of raster if extent does not match DEM.')
parser.add_argument('--noresample', dest='noresample', required=False, action='store_true',
                    help='Do not resample raster if its resolution differs from DEM. Will still resample if raster is not in the same spatial reference as DEM.')
parser.add_argument('-b', '--publisher', dest='publisher', required=False,
                    help="The publisher of the raster, if not supplied 'SELF PUBLISHED' will be used")
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if not context.config.has_option('GDAL/OGR', 'PATH_OF_GDAL_WARP'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_GDAL_WARP'))
    
if not context.config.has_option('GDAL/OGR', 'PATH_OF_GDAL_TRANSLATE'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_GDAL_TRANSLATE'))

if not os.access(args.rasterfile, os.R_OK):
    raise IOError(errno.EACCES, "Not allowed to read input raster %s" %
                  args.rasterfile)
inRasterPath = os.path.abspath(args.rasterfile)

if args.publisher:
    publisher = args.publisher
else:
    publisher = 'SELF PUBLISHED'

if args.outfile:
    outfile = args.outfile
else:
    outfile = args.type

force = False
if args.force:
    force = True

# Get study area parameters
studyArea = GenericMetadata.readStudyAreaEntries(context)
bbox = bboxFromString(studyArea['bbox_wgs84'])
demResolutionX = float(studyArea['dem_res_x'])
demResolutionY = float(studyArea['dem_res_y'])
demColumns = int(studyArea['dem_columns'])
demRows = int(studyArea['dem_rows'])
srs = studyArea['dem_srs']

rasterFilename = "%s%stif" % (outfile, os.extsep)
# Overwrite raster if already present
rasterFilepath = os.path.join(context.projectDir, rasterFilename)
if os.path.exists(rasterFilepath):
    os.unlink(rasterFilepath)

# Determine whether we need to resample
resample = False
rasterMetadata = getSpatialReferenceForRaster(inRasterPath)
rasterSrs = rasterMetadata[5]
rasterX = float(rasterMetadata[0])
rasterY = float(rasterMetadata[1])
if (rasterSrs != srs):
    resample = True
elif (not args.noresample) and ( (rasterX != demResolutionX) or (rasterY != demResolutionY) ):
    resample = True
    
if resample:
    # Reproject raster, copying into project directory in the process
    processingNotes = "Resampling %s raster from %s to %s, spatial resolution (%.2f, %.2f) to (%.2f, %.2f)" % \
        (args.type, rasterSrs, srs, rasterX, rasterX,
         demResolutionX, demResolutionY) 
    sys.stdout.write( textwrap.fill("%s..." % (processingNotes,) ) )
    resampleRaster(context.config, context.projectDir, inRasterPath, rasterFilepath, \
                   s_srs=None, t_srs=srs, \
                   trX=demResolutionX, trY=demResolutionY, \
                   resampleMethod=args.resampleMethod)
else:
    # Copy the raster in to the project directory
    processingNotes = "Importing %s raster from %s without resampling" % (args.type, inRasterPath)
    sys.stdout.write( textwrap.fill("%s..." % (processingNotes,) ) )
    sys.stdout.flush()
    copyRasterToGeoTIFF(context.config, context.projectDir, inRasterPath, rasterFilename)
sys.stdout.write('done\n')

# Make sure extent of resampled raster is the same as the extent of the DEM
newRasterMetadata = getDimensionsForRaster(rasterFilepath)
if (newRasterMetadata[0] != demColumns) or (newRasterMetadata[1] != demRows):
    if args.type == GenericMetadata.RASTER_TYPE_STREAM_BURNED_DEM:
        # Extents to not match, roll back and bail out
        os.unlink(rasterFilepath)
        sys.exit(textwrap.fill("ERROR: Raster type %s must be the same extent as DEM" % 
                 (GenericMetadata.RASTER_TYPE_STREAM_BURNED_DEM,) ) )
    if not force:
        # Extents to not match, roll back and bail out
        os.unlink(rasterFilepath)
        sys.exit(textwrap.fill("ERROR: Extent of raster dataset %s does not match extent of DEM in project directory %s. Use --force to override.") %
                 (rasterFilename, context.projectDir))

# Write metadata
if GenericMetadata.RASTER_TYPE_LC == args.type:
    GenericMetadata.writeStudyAreaEntry(context, "landcover_type", "custom")

# Write provenance
asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = args.type
asset.dcIdentifier = rasterFilename
asset.dcSource = "file://%s" % (inRasterPath,)
asset.dcTitle = args.type
asset.dcPublisher = publisher
asset.dcDescription = cmdline
asset.processingNotes = processingNotes
asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)




"""@package ecohydrolib.util
    
@brief Catch-all location for miscellaneous utility functions

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os


def getAbsolutePathOfItem(item):
    """ Attempt to get absolute path of items that exist in the file system.
        For non-existant items, quote spaces.
    
        @param item String representing item
        @return String representing the absolute path of the item.  If the item
        is not an existing file, the item string will be returned, but any 
        spaces will be quoted
    """
    if os.path.exists(item):
        result = os.path.abspath(item)
    else:
        if item.find(' ') != -1:
            # If a non-path item has spaces in it, quote them
            result = '"' + item + '"'
        else:
            result = item
            
    return result

def isExecutable(filepath):
        """ Check if a path is an executable file
            @param filepath String representing the path
            @return True if the path represents an executable file
        """
        return os.path.isfile(filepath) and os.access(filepath, os.X_OK)

def getAbsolutePathOfExecutable(program):
    """ Return the absolute path of an executable by searching through the
        PATH environment variable. 
        
        @note Adapted from http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
         
        @param program String representing the program
        
        @return The absolute path of the program, or None if the program was
        not found
    """
    fpath, fname = os.path.split(program)
    if fpath:
        if isExecutable(program):
            return os.path.abspath(program)
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            executable = os.path.join(path, program)
            if isExecutable(executable):
                return executable

    return None


"""@package ecohydrolib.wcslib
    
@brief Make WCS query for a raster data set

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os, errno
import sys

import socket
import httplib
import textwrap

from ecohydrolib.spatialdata.utils import deleteGeoTiff

CONTENT_TYPE_ERRORS = ['text/xml', 'application/vnd.ogc.se_xml;charset=ISO-8859-1']

_DEFAULT_CRS = 'EPSG:4326'
_BUFF_LEN = 4096 * 10

def getRasterForBoundingBox(config, outputDir, outFilename, host, urlProto, mimeType, bbox, coverage, srs, format, 
                            response_crs=None, store=None, resx=None, resy=None, interpolation=None, overwrite=True):
    """ Fetch a rater from WCS-compliant web service.
        Will write any error returned by query to sys.stderr.
    
        @param config A Python ConfigParser (not currently used)
        @param outputDir String representing the absolute/relative path of the directory into which output raster should be written
        @param outFilename String representing the name of the raster file to be written
        @param host String representing the host (e.g. 'webmap.ornl.gov', 'geobrain.laits.gmu.edu')
        @param urlProto String representing WCS service URL, must contain the following replacement fields:
            coverage, crs, bbox, format. May have the following fields: response_srs, store, resx, resy, interpolation
        @param mimeType String representing the MIME type expected for the response data
        @param bbox Dict representing the lat/long coordinates and spatial reference of the bounding box area
            for which the raster is to be extracted.  The following keys must be specified: minX, minY, maxX, maxY, srs.
        @param coverage String representing the raster source from which to get the raster coverage.  Must be a value listed in SUPPORTED_COVERAGE
        @param srs String representing the spatial reference of the raster to be returned.
        @param format String representing the MIME type of the raster format to be returned.  Must be a value listed in 
            SUPPORTED_FORMATS
        @param overwrite Boolean value indicating whether or not the file indicated by filename should be overwritten.
            If False and filename exists, IOError exception will be thrown with errno.EEXIST
    
        @raise IOError if outputDir is not a writable directory
        @raise IOError if outFilename already exists and overwrite is False (see above)
    
        @return Tuple(True if raster data were fetched and False if not, URL of raster fetched)
    """
    dataFetched = False
    assert(format)
    assert(coverage)
    assert('minX' in bbox)
    assert('minY' in bbox)
    assert('maxX' in bbox)
    assert('maxY' in bbox)
    assert('srs' in bbox)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    outFilepath = os.path.join(outputDir, outFilename)
    
    if os.path.exists(outFilepath):
        if overwrite: 
            deleteGeoTiff(outFilepath)
        else:
            raise IOError(errno.EEXIST, "Raster file %s already exists" % outFilepath)
    
    crs = bbox['srs']
    bboxStr = "%f,%f,%f,%f" % (bbox['minX'], bbox['minY'], bbox['maxX'], bbox['maxY'])
 
    # coverage, crs, bbox, format. May have the following fields: response_srs, store, resx, resy, interpolation
    url = urlProto.format(coverage=coverage, crs=crs, bbox=bboxStr, format=format, 
                          response_crs=srs, store=store, resx=resx, resy=resy, interpolation=interpolation)
    urlFetched = "http://%s%s" % (host, url)

    conn = httplib.HTTPConnection(host)
    try:
        conn.request('GET', url)
        res = conn.getresponse(buffering=True)
    except socket.error as e:
        msg = "Encountered the following error when trying to read raster from %s. Error: %s.  Please try again later or contact the developer." % \
            (urlFetched, str(e) )
        sys.stderr.write( textwrap.fill(msg) )
        return ( dataFetched, urlFetched )
  
    if 200 != res.status:
        msg = "HTTP response %d %s encountered when querying %s.  Please try again later or contact the developer." % \
            (res.status, res.reason, urlFetched)
        sys.stderr.write( textwrap.fill(msg) ) 
        return ( dataFetched, urlFetched )
     
    contentType = res.getheader('Content-Type')
    
    if contentType == mimeType:
        # The data returned were of the type expected, read the data
        data = res.read(_BUFF_LEN)
        if data: 
            demOut = open(outFilepath, 'wb')
            dataFetched = True
            while data:
                demOut.write(data)
                data = res.read(_BUFF_LEN)
            demOut.close()
    elif contentType in CONTENT_TYPE_ERRORS:
        # Read the error and print to stderr
        msg = "The following error was encountered reading URL %s\n" %\
            (urlFetched, )
        sys.stderr.write( textwrap.fill(msg) )
        data = res.read(_BUFF_LEN)
        while data: 
            sys.stderr.write(data)
            sys.stderr.flush()
            data = res.read(_BUFF_LEN)
        sys.stderr.write('\n')
    else:
        msg = "Query for raster from URL %s returned content type %s, was expecting type %s.  Operation failed." % \
            (urlFetched, contentType, mimeType)
        sys.stderr.write( textwrap.fill(msg) )
        
    return ( dataFetched, urlFetched )


"""@package ecohydrolib.spatialdata.utils
    
@brief Generic utilities for manipulating spatial data sets.
@brief Builds a task-oriented API, for select operations, on top of
GDAL/OGR utilities, GDAL/OGR API, Proj API.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>

@todo Refactor raster and vector functions into their own sub-packages
@toto Refactor bounding box as class
"""
import os, sys, errno
from math import sqrt
import math
import re

from osgeo.gdalconst import *
from osgeo import gdal
from osgeo import ogr
from osgeo import osr
from pyproj import Proj
from pyproj import transform
from pyproj import Geod

from shapely.geometry import shape

SHP_MINX = 0
SHP_MAXX = 1
SHP_MINY = 2
SHP_MAXY = 3

RASTER_RESAMPLE_METHOD = ['near', 'bilinear', 'cubic', 'cubicspline', 'lanczos', 'average', 'mode']
WGS84_EPSG = 4326
WGS84_EPSG_STR = 'EPSG:4326'

OGR_SHAPEFILE_DRIVER_NAME = 'ESRI Shapefile'
OGR_GEOJSON_DRIVER_NAME = 'GeoJSON'
OGR_DRIVERS = {OGR_SHAPEFILE_DRIVER_NAME: 'shp', 
               OGR_GEOJSON_DRIVER_NAME: 'geojson'}

EPSG_RE = re.compile('^epsg:\d+$')

def _readImageGDAL(filePath):
    ds = gdal.Open(filePath, GA_ReadOnly)
    cols = ds.RasterXSize
    rows = ds.RasterYSize
    trans = ds.GetGeoTransform()
    proj = ds.GetProjection()
    ds=None
    return { 'rows': rows, 'cols': cols, 'trans': trans, 'srs': proj }


def bboxFromString(bboxStr):
    """ Get bbox dictionary from comma separated string of the form
        '-76.769782 39.273610 -76.717498 39.326008'
    
        @param bboxStr String representing bounding box in WGS84 coordinates
        @return Dict representing bounding box with keys: minX, minY, maxX, maxY, and srs
    """
    bbox = bboxStr.split()
    bbox = dict({'minX': float(bbox[0]), 'minY': float(bbox[1]), 'maxX': float(bbox[2]), 'maxY': float(bbox[3]), 'srs': 'EPSG:4326'})
    return bbox


def isValidSrs(srsString):
    """ Determine if spatial reference string is valid.  Currently,
        only strings of the form 'EPSG:X...X' are valid.
        
        @param srsString
        
        @return True if valid, False if invalid
    """   
    m = EPSG_RE.match(srsString.lower())
    return m != None
    

def getEPSGStringForUTMZone(zone, isNorth):
    """ Get EPSG string, e.g. "EPSG:32618" for UTM zone (WGS84)
    
        @param zone Integer representing UTM zone
        @param isNorth True if north
        
        @return String of the pattern "^EPSG:\d+$"
    """
    if isNorth:
        epsg = 32600 + zone
    else:
        epsg = 32700 + zone
    return "EPSG:%d" % (epsg,)


def getUTMZoneFromCoordinates(longitude, latitude):
    """ Determine the UTM zone for coordinate pair
    
        @param longitude Float representing WGS84 longitude
        @param latitude Float representing WGS84 latitude
        
        @return Tuple of the form (zone number, true if north)
    """
    zone = int((math.floor((longitude + 180)/6) + 1) % 60)
    isNorth = False
    if latitude > 0:
        isNorth = True
    return (zone, isNorth)


def transformCoordinates(sourceX, sourceY, t_srs, s_srs="EPSG:4326"):
    """ Transform a pair of X,Y coordinates from one reference system to another
    
        @param sourceX A float representing the X coordinate
        @param sourceY A float representing the Y coordinate
        @param t_srs A string representing the spatial reference system of the output coordinates
        @param s_srs A string representing the spatial reference system, in EPSG format, of the input
            coordinates
        
        @return A tuple of floats representing the transformed coordinates
    """
    p_in = Proj(init=s_srs)
    p_out = Proj(init=t_srs)
    return transform(p_in, p_out, sourceX, sourceY)


def getRasterExtentAsBbox(extentRasterFilepath):
    """ Determine raster extent and store it in a bbox along with the spatial 
        reference system of the raster
    
        @param extentRasterFilepath String representing the path of the raster
        whose extent we are interested in
        @return Dict representing bounding box with keys: minX, minY, maxX, maxY, and srs
        
        @raise IOError(errno.ENOENT) if extentRasterFilepath does not exist
        @raise IOError(errno.EACCESS) if extentRasterFilepath is not readable
    """
    extentRasterFilepath = os.path.abspath(extentRasterFilepath)
    if not os.path.exists(extentRasterFilepath):
        raise IOError(errno.ENOENT, "Raster %s does not exist" % (extentRasterFilepath,) )
    if not os.access(extentRasterFilepath, os.R_OK):
        raise IOError(errno.EACCES, "Raster %s is not readable" % (extentRasterFilepath,) )
    
    extentImg = _readImageGDAL(extentRasterFilepath)
    
    rows = extentImg['rows']
    cols = extentImg['cols']
    targetResX = abs(extentImg['trans'][1])
    targetResY = abs(extentImg['trans'][5])
    xmin = extentImg['trans'][0]
    ymax = extentImg['trans'][3]
    xmax = xmin + (targetResX * cols)
    ymin = ymax - (targetResY * rows)
    
    # TODO: Refactor: This is inefficient because we are reading raster twice
    # But _readImageGDAL gives us SRS in WKT, whereas getSpatialReferenceForRaster
    # give us SRS in EPSG, which is what we need.
    srs = getSpatialReferenceForRaster(extentRasterFilepath)[5]
    
    bbox = dict({'minX': float(xmin), 'minY': float(ymin), 'maxX': float(xmax), 'maxY': float(ymax), 'srs': srs})
    return bbox


def extractTileFromRasterByRasterExtent(config, outputDir, extentRasterFilepath, inRasterFilepath, outRasterFilename, resampleMethod='near'):
    """ Extract a tile from a raster using the extent of another raster as the tile bounds.
        
        @note Output raster will be in LZW-compressed GeoTIFF file.
        @note Will delete existing output raster if it already exists, then perform the extraction
        
        @param config Python ConfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_GDAL_WARP'
        @param outputDir String representing the absolute/relative path of the directory into which output raster
         should be written
        @param extentRasterFilepath String representing the path of the extent raster
        @param inRasterFilepath String representing the path of the raster from which a tile will be extracted
        @param outRasterFilename String representing the name of the output raster
        @param resampleMethod String representing method to use to resample; one of: RASTER_RESAMPLE_METHOD
    """
    assert(resampleMethod in RASTER_RESAMPLE_METHOD)
    gdalCmdPath = config.get('GDAL/OGR', 'PATH_OF_GDAL_WARP')
    if not os.access(gdalCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "The gdalwarp binary at %s is not executable" %
                      gdalCmdPath)
    gdalCmdPath = os.path.abspath(gdalCmdPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    extentRasterFilepath = os.path.abspath(extentRasterFilepath)
    inRasterFilepath = os.path.abspath(inRasterFilepath)
    
    outRasterFilepath = os.path.join(outputDir, outRasterFilename)
    outRasterFilepath = os.path.abspath(outRasterFilepath)

    if os.path.exists(outRasterFilepath):
        os.unlink(outRasterFilepath)
    
    extentImg = _readImageGDAL(extentRasterFilepath)
    t_srs = extentImg['srs']
    inImg = _readImageGDAL(inRasterFilepath)
    s_srs = inImg['srs']
    
    rows = extentImg['rows']
    cols = extentImg['cols']
    targetResX = abs(extentImg['trans'][1])
    targetResY = abs(extentImg['trans'][5])
    xmin = extentImg['trans'][0]
    ymax = extentImg['trans'][3]
    xmax = xmin + (targetResX * cols)
    ymin = ymax - (targetResY * rows)
    
    targetExtent = [xmin, ymin, xmax, ymax]
    targetExtent = [str(x) for x in targetExtent]
    targetExtent = ' '.join(targetExtent)
    
    targetRes = "%f %f" % (targetResX, targetResY)
    
    gdalCommand = "%s -s_srs %s -t_srs %s -te %s -tr %s -r %s %s %s" \
        % (gdalCmdPath, s_srs, t_srs, targetExtent, targetRes, resampleMethod, inRasterFilepath, outRasterFilepath)
    #print gdalCommand
    returnCode = os.system(gdalCommand)
    if returnCode != 0:
        raise Exception("GDAL command %s failed." % (gdalCommand,)) 
            

def extractTileFromRaster(config, outputDir, inRasterFilename, outRasterFilename, bbox):
    """ Extract a tile from a raster. Tile extent is defined by supplied bounding box 
        with coordinates defined in WGS84 (EPSG:4326).
        
        @note Output raster will be in LZW-compressed GeoTIFF file.
        @note Will silently return if output raster already exists.
        
        @param config Python ConfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_GDAL_TRANSLATE'
        @param outputDir String representing the absolute/relative path of the directory into which output raster
         should be written
        @param inRasterFilename String representing the name of the input raster
        @param outRasterFilename String representing the name of the output raster
        @param bbox A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326' (WGS84)
    """
    gdalCmdPath = config.get('GDAL/OGR', 'PATH_OF_GDAL_TRANSLATE')
    if not os.access(gdalCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "The gdal_translate binary at %s is not executable" %
                      gdalCmdPath)
    gdalCmdPath = os.path.abspath(gdalCmdPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    inRasterFilepath = os.path.join(outputDir, inRasterFilename)
    outRasterFilepath = os.path.join(outputDir, outRasterFilename)
    
    if not os.path.exists(outRasterFilepath):
        # Convert into coordinate system of inRaster
        inRasterSrs = getSpatialReferenceForRaster(inRasterFilepath)
        inSrs = osr.SpatialReference()
        assert(inSrs.ImportFromWkt(inRasterSrs[4]) == 0)
        p_in = Proj(init="EPSG:4326")
        p_out = Proj(inSrs.ExportToProj4())
        (ulX, ulY) = transform(p_in, p_out, bbox['minX'], bbox['maxY'])
        (lrX, lrY) = transform(p_in, p_out, bbox['maxX'], bbox['minY'])
        
        gdalCommand = "%s -q -stats -of GTiff -co 'COMPRESS=LZW' -projwin %f %f %f %f %s %s" \
                        % (gdalCmdPath, ulX, ulY, lrX, lrY, \
                           inRasterFilepath, outRasterFilepath)
        #print gdalCommand
        returnCode = os.system(gdalCommand)
        if returnCode != 0:
            raise Exception("GDAL command %s failed." % (gdalCommand,)) 


def resampleRaster(config, outputDir, inRasterFilepath, outRasterFilename, \
                s_srs, t_srs, trX, trY, resampleMethod='bilinear'):
    """ Resample raster from one spatial reference system and resolution to another.
        
        @note Output raster will be in LZW-compressed GeoTIFF file.
        @note Will silently return if output raster already exists.
    
        @param config Python ConfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_GDAL_WARP'
        @param outputDir String representing the absolute/relative path of the directory into which output raster
            should be written
        @param inRasterFilepath String representing the path of the input raster
        @param outRasterFilename String representing the name of the output raster
        @param s_srs String representing the spatial reference of the input raster, if s_srs is None, 
            the input raster's spatial reference
        @param t_srs String representing the spatial reference of the output raster
        @param trX Float representing the X resolution of the output raster (in target spatial reference units)
        @param trY Float representing the Y resolution of the output raster (in target spatial reference units) 
        @param resampleMethod String representing resampling method to use. Must be one of spatialdatalib.utils.RASTER_RESAMPLE_METHOD.
        
        @exception ConfigParser.NoSectionError
        @exception ConfigParser.NoOptionError
        @exception IOError(errno.ENOTDIR) if outputDir is not a directory
        @exception IOError(errno.EACCESS) if outputDir is not writable
        @exception ValueError if trX or trY are not floating point numbers greater than 0
        @exception Exception if a gdal_warp command fails
    """
    gdalCmdPath = config.get('GDAL/OGR', 'PATH_OF_GDAL_WARP')
    if not os.access(gdalCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "The gdalwarp binary at %s is not executable" %
                      gdalCmdPath)
    gdalCmdPath = os.path.abspath(gdalCmdPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    trX = float(trX)
    if trX <= 0.0:
        raise ValueError("trX must be > 0.0")
    trY = float(trY)
    if trY <= 0.0:
        raise ValueError("trY must be > 0.0")
    
    assert(resampleMethod in RASTER_RESAMPLE_METHOD)
    
    nodata = getNodataValuesForRaster(inRasterFilepath)
    nodataStr = ' '.join(map(str, nodata))
    
    outRasterFilepath = os.path.join(outputDir, outRasterFilename)
    
    if not os.path.exists(outRasterFilepath):
        if s_srs is None:
            gdalCommand = "%s -q -t_srs %s -dstnodata '%s' -tr %f %f -r %s -of GTiff -co 'COMPRESS=LZW' %s %s" \
                            % (gdalCmdPath, t_srs, nodataStr, trX, trY, resampleMethod, \
                               inRasterFilepath, outRasterFilepath)
        else:
            gdalCommand = "%s -q -s_srs %s -t_srs %s -dstnodata '%s' -tr %f %f -r %s -of GTiff -co 'COMPRESS=LZW' %s %s" \
                            % (gdalCmdPath, s_srs, t_srs, nodataStr, trX, trY, resampleMethod, \
                               inRasterFilepath, outRasterFilepath)
        #print gdalCommand
        returnCode = os.system(gdalCommand)
        if returnCode != 0:
            raise Exception("GDAL command %s failed." % (gdalCommand,)) 


def convertGMLToShapefile(config, outputDir, gmlFilepath, layerName, t_srs):
    """ Convert a GML file to a shapefile.  Will silently exit if shapefile already exists
    
        @param config A Python ConfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_OGR2OGR'
        @param outputDir String representing the absolute/relative path of the directory into which shapefile should be written
        @param gmlFilepath String representing the absolute path of the GML file to convert
        @param layerName String representing the name of the layer contained in the GML file to write to a shapefile
        @param t_srs String representing the spatial reference system of the output shapefile, of the form 'EPSG:XXXX'
        
        @return String representing the name of the shapefile written
    
        @exception Exception if the conversion failed.
    """
    pathToOgrCmd = config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    shpFilename = "%s.shp" % (layerName,)
    shpFilepath = os.path.join(outputDir, shpFilename)
    if not os.path.exists(shpFilepath):
        ogrCommand = "%s -f 'ESRI Shapefile' -nln %s -t_srs %s %s %s" % (pathToOgrCmd, "MapunitPolyExtended", t_srs, shpFilepath, gmlFilepath)
        returnCode = os.system(ogrCommand)
        if returnCode != 0:
            raise Exception("GML to shapefile command %s returned %d" % (ogrCommand, returnCode))
    
    return shpFilename


def convertGMLToGeoJSON(config, outputDir, gmlFilepath, layerName, t_srs='EPSG:4326'):
    """ Convert a GML file to a shapefile.  Will silently exit if GeoJSON already exists
    
        @param config A Python ConfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_OGR2OGR'
        @param outputDir String representing the absolute/relative path of the directory into which GeoJSON should be written
        @param gmlFilepath String representing the absolute path of the GML file to convert
        @param layerName String representing the name of the layer contained in the GML file to write to a GeoJSON
        @param t_srs String representing the spatial reference system of the output GeoJSON, of the form 'EPSG:XXXX'
        
        @return String representing the name of the GeoJSON written
    
        @exception Exception if the conversion failed.
    """
    pathToOgrCmd = config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    geojsonFilename = "%s.geojson" % (layerName,)
    geojsonFilepath = os.path.join(outputDir, geojsonFilename)
    if not os.path.exists(geojsonFilepath):
        ogrCommand = "%s -f 'GeoJSON' -nln %s -t_srs %s %s %s" % (pathToOgrCmd, layerName, t_srs, geojsonFilepath, gmlFilepath)
        returnCode = os.system(ogrCommand)
        if returnCode != 0:
            raise Exception("GML to GeoJSON command %s returned %d" % (ogrCommand, returnCode))
    
    return geojsonFilename


def convertGeoJSONToShapefile(config, outputDir, geoJSONFilepath, shapefileName, layerName='OGRGeoJSON', t_srs='EPSG:4326'):
    """ Convert a GeoJSON file to a shapefile.  Will silently exit if shapefile already exists
    
        @param config A Python ConfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_OGR2OGR'
        @param outputDir String representing the absolute/relative path of the directory into which shapefile should be written
        @param geoJSONFilepath String representing the absolute path of the GeoJSON file to convert
        @param layerName String representing the name of the layer contained in the GeoJSON file to write to a shapefile
        @param t_srs String representing the spatial reference system of the output shapefile, of the form 'EPSG:XXXX'
        
        @return String representing the name of the shapefile written
    
        @exception Exception if the conversion failed.
    """
    pathToOgrCmd = config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    shpFilename = "%s.shp" % (shapefileName,)
    shpFilepath = os.path.join(outputDir, shpFilename)
    if not os.path.exists(shpFilepath):
        ogrCommand = "%s -f 'ESRI Shapefile' -nln %s -t_srs %s %s %s" % (pathToOgrCmd, layerName, t_srs, shpFilepath, geoJSONFilepath)
        returnCode = os.system(ogrCommand)
        if returnCode != 0:
            raise Exception("GeoJSON to shapefile command %s returned %d" % (ogrCommand, returnCode))
    
    return shpFilename


def convertFeatureLayerToShapefile(config, outputDir, featureFilepath, shapefileName, layerName=None, t_srs='EPSG:4326', overwrite=False):
    """ Convert a vector feature file readible by OGR to a shapefile.  
        Will silently exit if output shapefile already exists
    
        @param config A Python ConfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_OGR2OGR'
        @param outputDir String representing the absolute/relative path of the directory into which shapefile should be written
        @param featureFilepath String representing the absolute path of the feature file to convert
        @param layerName String representing the name of the layer contained in the feature file to write to a shapefile.
                         If none, the base name of featureFilepath will be used
        @param t_srs String representing the spatial reference system of the output shapefile, of the form 'EPSG:XXXX'
        
        @return String representing the name of the shapefile written
    
        @exception Exception if the conversion failed.
    """
    pathToOgrCmd = config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    if not os.access(featureFilepath, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read feature path %s" % (featureFilepath,))
    
    if layerName == None:
        layerName = os.path.splitext(os.path.basename(featureFilepath))[0]
    
    shpFilename = "%s.shp" % (shapefileName,)
    shpFilepath = os.path.join(outputDir, shpFilename)
    
    if os.path.exists(shpFilepath):
        if overwrite:
            deleteShapefile(shpFilepath)
        else:
            raise Exception("Output shapefile %s already exists, but the overwrite option was not specified." % (shpFilepath,) )
        
    # Import the feature layer using ogr2ogr...
    ogrCommand = "%s -f 'ESRI Shapefile' -nln %s -t_srs %s %s %s" % (pathToOgrCmd, layerName, t_srs, shpFilepath, featureFilepath)
    returnCode = os.system(ogrCommand)
    if returnCode != 0:
        raise Exception("Feature layer to shapefile command %s returned %d" % (ogrCommand, returnCode))
    
    return shpFilename


def deleteShapefile(shpfilePath):
    """ Delete shapefile and its related files (.dbf, .prj, .shx)
        
        @param shpfilePath -- Path, including filename, of the shapefile to be deleted
    """
    SHP_EXT = ['dbf', 'prj', 'shx']
    fileName = os.path.splitext(shpfilePath)[0]
    if os.path.exists(shpfilePath):
        os.remove(shpfilePath)
    for ext in SHP_EXT:
        tmpFilepath = fileName + "." + ext
        if os.path.exists(tmpFilepath):
            os.remove(tmpFilepath)


def getCoordinatesOfPointsFromShapefile(shpFilepath, layerName, pointIDAttr, pointIDs):
    """ Get WGS84 coordinates of point features in shapefile
    
        @param shpFilepath String representing the path of the shapefile
        @param layerName String representing the name of the layer within the shapefile from 
        to read points
        @param pointIDAttr String representing name of the attribute used to identify points
        @param pointIDs List of strings representing IDs of coordinate pairs
        
        @raise Exception if unable to read layer of shapefile, or if no feature(s) with given
        attribute value(s) were found in the shapefile
        
        @return Tuple of floats of the form (longitude, latitude)
    """
    coordinates = []
    
    if not os.access(shpFilepath, os.R_OK):
        raise IOError(errno.EACCES, "Unable to read shapefile %s" % (shpFilepath,))
    
    poDS = ogr.Open(shpFilepath, True)
    assert(poDS.GetLayerCount() > 0)
    poLayer = poDS.GetLayerByName(layerName)
    if not poLayer:
        raise Exception( "Layer named '%s' not found in shapefile %s" % (layerName, shpFilepath) )
    
    # Determine type of ID field
    poFeatureDef = poLayer.GetLayerDefn()
    assert(poFeatureDef)
    idAttrType = ogr.OFTString
    numFields = poFeatureDef.GetFieldCount()
    i = 0
    while i < numFields:
        poFieldDef = poFeatureDef.GetFieldDefn(i)
        if pointIDAttr == poFieldDef.GetNameRef():
            idAttrType = poFieldDef.GetType()
            break
        i = i + 1
    
    # Build query string to select points of interest
    assert(len(pointIDs) > 0)
    if ogr.OFTString == idAttrType:
        whereFilter = "%s='%s'" % (pointIDAttr, pointIDs[0])
        for point in pointIDs[1:]:
            whereFilter = "%s and %s=%s" % (whereFilter, pointIDAttr, point)
    else:
        whereFilter = "%s=%s" % (pointIDAttr, pointIDs[0])
        for point in pointIDs[1:]:
            whereFilter = "%s and %s=%s" % (whereFilter, pointIDAttr, point)
    
    # Determine spatial reference, and if we need to convert coordinates
    isWGS84 = True
    inSRS = poLayer.GetSpatialRef()
    epsgStr = "%s:%s" % ( inSRS.GetAttrValue("AUTHORITY", 0), inSRS.GetAttrValue("AUTHORITY", 1) )
    if WGS84_EPSG_STR != epsgStr:
        p_in = Proj(inSRS.ExportToProj4())
        p_out = Proj(init="EPSG:4326")
        isWGS84 = False
    
    # Iterate over features matching query string
    if poLayer.SetAttributeFilter(whereFilter) != 0:
        raise Exception( "Feature identified by \"%s\" not found in layer '%s' of shapefile %s" % \
                         (whereFilter, layerName, shpFilepath) )
    poFeature = poLayer.GetNextFeature()
    while poFeature:
        poGeometry = poFeature.GetGeometryRef()
        
        # Get coordinates
        x = poGeometry.GetX()
        y = poGeometry.GetY()
        
        # Convert coorinate pair to WGS84, if need be
        if not isWGS84:
            (x, y) = transform(p_in, p_out, x, y)
        
        coordinates.append( (x, y) )
        poFeature = poLayer.GetNextFeature()
        
    if len(coordinates) == 0:
        raise Exception( "No features identified by \"%s\" not found in layer '%s' of shapefile %s" % \
                         (whereFilter, layerName, shpFilepath) )
    return coordinates


def deleteGeoTiff(geoTiffPath):
    """ Delete GeoTIFF and its related files (.aux.xml)
        
        @param geoTiffPath -- Path, including filename, of the GeoTIFF to be deleted
    """
    GTIFF_EXT = ['tif.aux.xml']
    fileName = os.path.splitext(geoTiffPath)[0]
    if os.path.exists(geoTiffPath):
        os.remove(geoTiffPath)
    for ext in GTIFF_EXT:
        tmpFilepath = fileName + "." + ext
        if os.path.exists(tmpFilepath):
            os.remove(tmpFilepath)


def isCoordinatePairInBoundingBox(bbox, coordinates):
    """ Determine whether coordinate pair lies within bounding box
    
        @param bbox A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
        @param coordinates List of tuples of floats of the form (longitude, latitude), in WGS84
    
        @return True if coordinates pair is within bounding box
    """
    assert(bbox['srs'] == 'EPSG:4326')
    
    lon = coordinates[0]
    lat = coordinates[1]
    
    if (lon < bbox['minX']) or (lon > bbox['maxX']):
        return False 
    if (lat < bbox['minY']) or (lat > bbox['maxY']):
        return False
    
    return True


def calculateBoundingBoxCenter(bbox):
    """ Calculate the central point of the bounding box
    
        @param bbox A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
        
        @return Tuple of floats of the form (longitude, latitude)
    """
    x_diff = ( bbox['maxX'] - bbox['minX'] ) / 2
    y_diff = ( bbox['maxY'] - bbox['minY'] ) / 2
    longitude = bbox['minX'] + x_diff
    latitude = bbox['minY'] + y_diff
    return (longitude, latitude)


def calculateBoundingBoxArea(bbox, srs='EPSG:4326'):
    """ Calculate bbox area in squared units of srs
    
        @param bbox A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
        @param srs String representing spatial reference system, in EPSG format, in which
        area should be calculated.
        
        @return Float representing the bounding box area in square meters
    """
    assert(bbox['srs'] == WGS84_EPSG_STR)

    coords = [(bbox['minX'], bbox['minY']), 
              (bbox['maxX'], bbox['minY']),
              (bbox['maxX'], bbox['maxY']), 
              (bbox['minX'], bbox['maxY'])]
    (lon, lat) = zip(*coords)
    p_in = Proj(init=bbox['srs'])
    p_out = Proj(init=srs)
    (x, y) = transform(p_in, p_out, lon, lat)
    geojson = {'type': 'Polygon', 'coordinates': [zip(x, y)]}

    return shape(geojson).area


def tileBoundingBox(bbox, threshold):
    """ Break up bounding box into tiles if bounding box is larger than threshold. 
        Bounding box must be defined by WGS84 lat,lon coordinates
        
        @param bbox Dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
        @param threshold Float representing threshold area above which bounding box will be tiled. Units: sq. meters
        
        @return A list containing tiles defined as a dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
    """
    assert(bbox['srs'] == 'EPSG:4326')
    area = calculateBoundingBoxArea(bbox)
    
    bboxes = []
    if area <= threshold:
        sys.stderr.write("Area of bounding box <= threshold, no tiling\n")
        bboxes.append(bbox)
    else:
        sys.stderr.write("Area of bounding box > threshold, tiling ...\n")
        # Tile it
        # Calculate the length of a "side" of a square tile
        tileSide = sqrt(threshold)
        # Start at the southwestern-most corner of the bounding box
        minLat = bbox['minY']
        minLon = bbox['minX']
        #maxLat = None; maxLon = None
        while minLat < bbox['maxY']: 
            (lon, maxLat, backAz) = geod.fwd(lons=minLon, lats=minLat, az=NORTH, dist=tileSide)
            while minLon < bbox['maxX']:
                (maxLon, lat, backAz) = geod.fwd(lons=minLon, lats=maxLat, az=EAST, dist=tileSide)
                bboxes.append(dict({'minX': minLon, 'minY': minLat, 'maxX': maxLon, 'maxY': maxLat, 'srs': 'EPSG:4326'}))
                minLon = maxLon
            minLat = maxLat
            minLon = bbox['minX']
                                                
    return bboxes


def getBoundingBoxForShapefile(shapefileName, buffer=0.0):
    """ Return the bounding box, in WGS84 (EPSG:4326) coordinates, for the ESRI shapefile.  
        Assumes shapefile exists and is readable.
        Based on http://svn.osgeo.org/gdal/trunk/gdal/swig/python/samples/ogrinfo.py
   
        @param shapefileName String representing the path of the shapefile whose bounding box should be determined.
        @param buffer Float >= 0.0 representing number of degrees by which to buffer the bounding box; 0.0 = no buffer, 
        0.01 = 0.01 degree buffer
        
        @return A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
    """
    assert(buffer >= 0.0)
    
    # Get spatial reference system for shapefile
    poDS = ogr.Open(shapefileName, True)
    assert(poDS.GetLayerCount() > 0)
    poLayer = poDS.GetLayer(0)
    assert(poLayer)
    srs_proj4 = poLayer.GetSpatialRef().ExportToProj4()
    
    # Setup Proj to convert to EPSG:4326 (WGS84)
    p_in = Proj(srs_proj4)
    p_out = Proj(init="EPSG:4326")
    
    # Get bounding box for shapefile
    (minX, maxX, minY, maxY) = poLayer.GetExtent()
    # Convert coordinates to EPSG:4326 (WGS84)
    (minX, minY) = transform(p_in, p_out, minX, minY)
    (maxX, maxY) = transform(p_in, p_out, maxX, maxY)
    
    bbox = dict({'minX': minX, 'minY': minY, 'maxX': maxX, 'maxY': maxY, 'srs': 'EPSG:4326'})
    bufferBoundingBox(bbox, buffer)
    
    return bbox


def bufferBoundingBox(bbox, buffer):
    """ Buffer the bounding by a given percentage
    
        @param bbox A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
        @param buffer Float >= 0.0 representing number of degrees by which to buffer the bounding box; 0.0 = no buffer, 
        0.01 = 0.01 degree buffer
    """
    if buffer > 0.0:
        # Apply buffer
        minX = bbox['minX']
        minY = bbox['minY']
        maxX = bbox['maxX']
        maxY = bbox['maxY']
        minX = minX - buffer
        minY = minY - buffer
        maxX = maxX + buffer
        maxY = maxY + buffer
        # Correct any wrapping that occurred
        if minX < -180.0: minX = 360.0 + minX
        if minY < -90.0: minY = -90.0
        if maxX > 180.0: maxX = -360.0 + maxX
        if minY > 90.0: maxY = 90.0
        bbox['minX'] = minX
        bbox['minY'] = minY
        bbox['maxX'] = maxX
        bbox['maxY'] = maxY


def getMeterConversionFactorForLinearUnitOfGMLfile(gmlFilename):
    """ Get conversion factor for converting a GML file's linear unit into meters
    
        @param gmlFilename String representing the GML file
        
        @return Float representing the conversion factor
    
        @exception IOError(errno.EACCES) if the GML file cannot be opened
    """
    driver = ogr.GetDriverByName('GML')
    inDS = driver.Open(gmlFilename, 0)
    if inDS is None:
        raise IOError(errno.EACCES, "Unable to open GML file %s" %
                      gmlFilename)
    inLayer = inDS.GetLayer()
    soilSrs = inLayer.GetSpatialRef()
    if soilSrs == None:
        # Assume EPSG:4326 (WGS84 GCS)
        soilSrs = osr.SpatialReference()
        soilSrs.ImportFromEPSG(4326)
    return soilSrs.GetLinearUnits()


def getMeterConversionFactorForLinearUnitOfShapefile(shpFilename):
    """ Get conversion factor for converting a shapefile's linear unit into meters
    
        @param shpFilename String representing the name of the shapefile
        
        @return Float representing the conversion factor
    
        @exception IOError(errno.EACCES) if the shapefile cannot be opened
    """
    driver = ogr.GetDriverByName('ESRI Shapefile')
    inDS = driver.Open(shpFilename, 0)
    if inDS is None:
        raise IOError(errno.EACCES, "Unable to open shapefile %s" %
                      shpFilename)
    inLayer = inDS.GetLayer()
    soilSrs = inLayer.GetSpatialRef()
    if soilSrs == None:
        # Assume EPSG:4326 (WGS84 GCS)
        soilSrs = osr.SpatialReference()
        soilSrs.ImportFromEPSG(4326)
    return soilSrs.GetLinearUnits()


def getNodataValuesForRaster(filename):
    """ Get nodata value for each band in a raster.  Uses GDAL library
        Code adapted from: http://svn.osgeo.org/gdal/trunk/gdal/swig/python/samples/gdalinfo.py
        
        @param filename String representing the DEM file path
        
        @return List containing nodata values for each band
        
        @exception IOError if filename is not readable
    """
    nodata = []
    
    if not os.access(filename, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read DEM %s to nodata value" %
                      filename)
        
    hDataset = gdal.Open(filename, gdal.GA_ReadOnly)
    if hDataset is not None:
        for iBand in range(hDataset.RasterCount):
            hBand = hDataset.GetRasterBand(iBand+1 )
            nodata.append( hBand.GetNoDataValue() )
    
    return nodata


def getSpatialReferenceForRaster(filename):
    """ Get pixel size and unit for DEM.  Uses GDAL library
        Code adapted from: http://svn.osgeo.org/gdal/trunk/gdal/swig/python/samples/gdalinfo.py
        
        @param filename String representing the DEM file path
        
        @return A tuple of the form: 
        (pixelWidth, pixelHeight, linearUnitsName, linearUnitsConversionFactor, WKT SRS string, EPSG SRS string)
        
        @exception IOError if filename is not readable
    """
    pixelWidth = None
    pixelHeight = None
    linearUnitsName = None
    linearUnitsConversionFactor = None
    pszProjection = None
    
    if not os.access(filename, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read DEM %s to read spatial reference" %
                      filename)
    
    hDataset = gdal.Open(filename, gdal.GA_ReadOnly)

    if hDataset is not None:
        adfGeoTransform = hDataset.GetGeoTransform()
        if adfGeoTransform is not None:
            pixelWidth = abs(adfGeoTransform[1])
            pixelHeight = abs(adfGeoTransform[5])
        
        pszProjection = hDataset.GetProjectionRef()
        if pszProjection is not None:
            hSRS = osr.SpatialReference()
            if hSRS.ImportFromWkt(pszProjection) == gdal.CE_None:
                linearUnitsName = hSRS.GetLinearUnitsName()
                linearUnitsConversionFactor = hSRS.GetLinearUnits()
        
        if hSRS.GetAttrValue("AUTHORITY", 0) != None and hSRS.GetAttrValue("AUTHORITY", 1):
            epsgStr = "%s:%s" % ( hSRS.GetAttrValue("AUTHORITY", 0), hSRS.GetAttrValue("AUTHORITY", 1) )
        else:
            epsgStr = None
    return (pixelWidth, pixelHeight, linearUnitsName, linearUnitsConversionFactor, pszProjection, epsgStr)


def getDimensionsForRaster(filename):
    """ Get number of columns and rows for raster.  Uses GDAL library
        Code adapted from: http://svn.osgeo.org/gdal/trunk/gdal/swig/python/samples/gdalinfo.py
        
        @param filename String representing the DEM file to read pixel size and units
        
        @return A tuple of the form: 
        (columns, rows) or None if raster could not be opened
        
        @exception IOError if filename is not readable
    """
    columns = None
    rows = None
    
    if not os.access(filename, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read DEM %s to read number of columns and rows" %
                      filename)
    
    hDataset = gdal.Open(filename, gdal.GA_ReadOnly)

    if hDataset is not None:
        columns = hDataset.RasterXSize
        rows = hDataset.RasterYSize
            
    return (columns, rows)


def getBoundingBoxForRaster(filename):
    """ Return the bounding box, in WGS84 (EPSG:4326) coordinates, for the raster dateset.  
        Assumes raster exists and is readable.
        Code adapted from: http://svn.osgeo.org/gdal/trunk/gdal/swig/python/samples/gdalinfo.py
        
        @param filename String representing the DEM file to read pixel size and units
        
        @return A tuple of the form: 
        (columns, rows) or None if raster could not be opened
        
        @exception IOError if filename is not readable
        @exception Exception if raster dataset failed to open
    """
    
    if not os.access(filename, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read DEM %s to determine bounding box" %
                      filename)
    
    hDataset = gdal.Open(filename, gdal.GA_ReadOnly)

    if hDataset is None:
        raise Exception("Unable to open raster dataset")
    
    # Setup for translating from pixels to coordinates
    pszProjection = hDataset.GetProjectionRef()
    assert(pszProjection is not None)
    hProj = osr.SpatialReference(pszProjection)
    assert(hProj is not None)
    hLatLong = osr.SpatialReference()
    hLatLong.ImportFromEPSG(WGS84_EPSG)
    hTransform = osr.CoordinateTransformation( hProj, hLatLong )
    adfGeoTransform = hDataset.GetGeoTransform(can_return_null = True)
    assert(adfGeoTransform is not None)

    (minX, maxY) = _transformPixelsToCoordinates(hDataset, hTransform, adfGeoTransform,
                                                 0, 0)
    (maxX, minY) = _transformPixelsToCoordinates(hDataset, hTransform, adfGeoTransform,
                                                 hDataset.RasterXSize, hDataset.RasterYSize)

    return dict({'minX': float(minX), 'minY': float(minY), 'maxX': float(maxX), 'maxY': float(maxY), 'srs': 'EPSG:4326'})


def writeBboxPolygonToShapefile(bbox, outputDir, layerName, overwrite=True):
    """ Write bbox to a shapfile
    
        @param bbox A dict containing keys: minX, minY, maxX, maxY, srs, where srs='EPSG:4326'
        @param outputDir String representing the absolute/relative path of the directory into which shapefile should be written
        @param layerName String representing the name of the layer. Will be used as root of filename of output shapefile
        @param overwrite True if existing shapefile should be overwritten.
        
        @return String representing the name of shapefile created (not the absolute path)
        
        @raise IOError is output directory is not a writable directory
        @raise Exception if shapefile already exists
        @raise Exception is failed to create shapefile
    """
    pszDriverName = "ESRI Shapefile"
    shpFilename = "%s%sshp" % (layerName, os.extsep)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    shpFilepath = os.path.join(outputDir, shpFilename)
    if os.path.exists(shpFilepath):
        if not overwrite:
            raise Exception("Shapefile %s already exists in directory %s" % \
                            (shpFilename,outputDir) )
        else:
            deleteShapefile(shpFilepath)
    
    ogr.UseExceptions()
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(WGS84_EPSG)
    hDriver = ogr.GetDriverByName(pszDriverName)
    assert(hDriver is not None)
    hDS = hDriver.CreateDataSource(shpFilepath)
    assert(hDS is not None)
    hLayer = hDS.CreateLayer(layerName, srs, ogr.wkbPolygon)
    assert(hLayer is not None)
    hFeature = ogr.Feature(hLayer.GetLayerDefn())
    ring = ogr.Geometry(type=ogr.wkbLinearRing)
    ring.AddPoint(bbox['minX'], bbox['minY'])
    ring.AddPoint(bbox['minX'], bbox['maxY'])
    ring.AddPoint(bbox['maxX'], bbox['maxY'])
    ring.AddPoint(bbox['maxX'], bbox['minY'])
    poly = ogr.Geometry(type=ogr.wkbPolygon)
    poly.AssignSpatialReference(srs)
    poly.AddGeometry(ring)
    hFeature.SetGeometry(poly)
    if hLayer.CreateFeature(hFeature) != 0:
        raise Exception("Failed to create shapefile for bounding box")
    # Clean-up
    hFeature.Destroy()
    hDS.Destroy()
    
    return shpFilename


def writeCoordinatePairsToPointShapefile(outputDir, layerName, pointIDAttr, pointIDs, coordinates, overwrite=True):
    """ Write coordinates as a point shapefile
    
        @param outputDir String representing the absolute/relative path of the directory into which shapefile should be written
        @param layerName String representing the name of the layer. Will be used as root of filename of output shapefile
        @param pointIDAttr String representing name of the attribute used to identify points
        @param pointIDs List of strings representing IDs of coordinate pairs 
        @param coordinates List of tuples of floats of the form (longitude, latitude), in WGS84
        @param overwrite True if existing shapefile should be overwritten.
        
        @return String representing the name of shapefile created (not the absolute path)
        
        @raise IOError is output directory is not a writable directory
        @raise Exception if shapefile already exists
        @raise Exception is failed to create shapefile
    """
    pszDriverName = "ESRI Shapefile"
    shpFilename = "%s%sshp" % (layerName, os.extsep)
    
    numCoord = len(coordinates)
    assert(len(pointIDs) == numCoord)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    shpFilepath = os.path.join(outputDir, shpFilename)
    if os.path.exists(shpFilepath):
        if not overwrite:
            raise Exception("Shapefile %s already exists in directory %s" % \
                            (shpFilename,outputDir) )
        else:
            deleteShapefile(shpFilepath)
    
    ogr.UseExceptions()
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(WGS84_EPSG)
    hDriver = ogr.GetDriverByName(pszDriverName)
    assert(hDriver is not None)
    hDS = hDriver.CreateDataSource(shpFilepath)
    assert(hDS is not None)
    hLayer = hDS.CreateLayer(layerName, srs, ogr.wkbPoint)
    assert(hLayer is not None)
    
    # Create ID field
    hField = ogr.FieldDefn(pointIDAttr, ogr.OFTString)
    hField.SetWidth(32)
    assert(hLayer.CreateField(hField) == 0)
    
    i = 0
    while i < numCoord:
        x = coordinates[i][0]
        y = coordinates[i][1]
        hFeature = ogr.Feature(hLayer.GetLayerDefn())
        hFeature.SetField(pointIDAttr, pointIDs[i])
        hPoint = ogr.Geometry(ogr.wkbPoint)
        hPoint.AssignSpatialReference(srs)
        hPoint.SetPoint_2D(0, x, y)
        hFeature.SetGeometry(hPoint)
        assert(hLayer.CreateFeature(hFeature) == 0)
        hFeature.Destroy()
        i = i + 1
    
    hDS.Destroy()
    
    return shpFilename


def copyRasterToGeoTIFF(config, outputDir, inRasterPath, outRasterName):
    """ Copy input raster from a location outside of outputDir to a GeoTIFF format raster stored in outputDir
    
        @param config A Python ConfigParser containing the section 'GDAL/OGR' and option 'PATH_OF_GDAL_TRANSLATE'
        @param outputDir String representing the absolute/relative path of the directory into which shapefile should be written
        @param inRasterPath String representing path of input raster
        @param outRasterName String representing name of output raster to be stored in outputDir
        
        @raise IOError if gdal_translate binary is not found/executable
        @raise IOError if output directory does not exist or not writable
        @raise IOError if input raster is not readable
    """
    gdalCmdPath = config.get('GDAL/OGR', 'PATH_OF_GDAL_TRANSLATE')
    if not os.access(gdalCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "The gdal_translate binary at %s is not executable" %
                      gdalCmdPath)
    gdalCmdPath = os.path.abspath(gdalCmdPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    if not os.access(inRasterPath, os.R_OK):
        raise IOError(errno.EACCES, "Not allowed to read input raster %s" (inRasterPath,))
    
    outRasterPath = os.path.join(outputDir, outRasterName)
    
    gdalCommand = "%s -q -of GTiff -co 'COMPRESS=LZW' %s %s" % \
                  (gdalCmdPath, inRasterPath, outRasterPath)
    returnCode = os.system(gdalCommand)
    if returnCode != 0:
        raise Exception("GDAL command %s failed." % (gdalCommand,)) 
    
    

def _transformPixelsToCoordinates(hDataset, hTransform, adfGeoTransform, x, y):
    """ Adapted from http://svn.osgeo.org/gdal/trunk/gdal/swig/python/samples/gdalinfo.py
    
        @param hDataset A GDAL raster dataset object
        @param hTransform A GDAL transform object
        @param adfGeoTransform A GDAL geographic transform object
        @param x The X coordinate
        @param y The Y coordinate
        @return Tuple of floats representing longitude and latitude coordiates.
    """
    # Transform point into georeferenced coordiates
    dfGeoX = adfGeoTransform[0] + adfGeoTransform[1] * x \
            + adfGeoTransform[2] * y
    dfGeoY = adfGeoTransform[3] + adfGeoTransform[4] * x \
            + adfGeoTransform[5] * y
    # Transform georeferenced coordinates into lat/long
    coords = hTransform.TransformPoint(dfGeoX, dfGeoY, 0)
    return (coords[0], coords[1])
        
    
    


#!/usr/bin/env python
"""@package DumpMetadataToiRODSXML

@brief Dump EcohydroLib metadata to iRODS XML metadata import format

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  

Pre conditions:
--------------
1. Metadata exist for the specified project directory
   

Post conditions:
----------------
2. A file named metadata.xml will be written to the specified project directory

Usage:
@code
DumpMetadataToiRODSXML.py -p /path/to/project_dir -c /irods/collection/path
@endcode
"""
import os
import codecs
import argparse
from xml.sax.saxutils import escape

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance


PATH_SEP_IRODS = '/'
OUTFILE_NAME = 'metadata.xml'

def writeAVUToXMLFile(outfile, target, attribute, value, unit=None):
    """ Write Attribute, Value, Unit (AVU) element to iRODS metadata XML file
    
        @param outfile StreamWriter representing the XML file. It is assumed that opening 
        <metadata> element has already been written to the file
        @param target String representing the contents of the Target element
        @param attribute String representing the contents of the Attribute element
        @param value String representing the contents of the Value element
        @param unit String representing the contents of the Unit element.  
        If None, an empty element will be written 
    """
    outfile.write('\t<AVU>\n')
    outfile.write("\t\t<Target>%s</Target>\n" % (escape(target),))
    outfile.write("\t\t<Attribute>%s</Attribute>\n" % (escape(attribute),) )
    outfile.write("\t\t<Value>%s</Value>\n" % (escape(value),) )
    if unit:
        outfile.write("\t\t<Unit>%s</Unit>\n" % (unit,) )
    else:
        outfile.write('\t\t<Unit />\n')
    outfile.write('\t</AVU>\n')

def writeDictToXMLFile(outfile, target, dict):
    """ Write the contents of a dict as AVU elements in an iRODS metadata XML file
    
        @param outfile StreamWriter representing the XML file. It is assumed that opening 
        <metadata> element has already been written to the file
        @param target String representing the contents of the Target element
        @param dict The dictionary whose key will serve as attributes and whose values will serve as
        values. Units will be written as empty elements for each AVU written
    """
    targetStr = "\t\t<Target>%s</Target>\n" % (escape(target),)
    for key in dict.keys():
        outfile.write('\t<AVU>\n')
        outfile.write(targetStr)
        outfile.write("\t\t<Attribute>%s</Attribute>\n" % (escape(key),) )
        outfile.write("\t\t<Value>%s</Value>\n" % (escape(dict[key]),) )
        outfile.write('\t\t<Unit />\n')
        outfile.write('\t</AVU>\n')


parser = argparse.ArgumentParser(description='Dump point climate station information from EcohydroLib metadata to standard output')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory from which metadata should be read')
parser.add_argument('-c', '--collection', dest='collection', required=True,
                    help='The iRODS collection corresponding to the project directory')
args = parser.parse_args()

context = Context(args.projectDir, None) 

# Make sure there's no trailing PATH_SEP_IRODS on the collection
collection = args.collection.rstrip(PATH_SEP_IRODS)

outfilePath = os.path.join(context.projectDir, OUTFILE_NAME)
outfile = codecs.getwriter('utf-8')(open(outfilePath, 'w')) 
outfile.write('<?xml version="1.0" encoding="UTF-8" ?>\n')
outfile.write('<metadata>\n')

# Write study area metadata to collection root
writeDictToXMLFile(outfile,  collection, GenericMetadata.readStudyAreaEntries(context))

# Write processing history to collection root
history = GenericMetadata.getProcessingHistoryList(context)
i = 1
for entry in history:
    attribute = "processing_step_%d" % (i,); i += 1
    writeAVUToXMLFile(outfile, collection, attribute, entry)

# Write provenance to each item in the manifest
provenance = GenericMetadata.readAssetProvenanceObjects(context)
for entry in provenance:
    target = collection + PATH_SEP_IRODS + entry.dcIdentifier
    writeAVUToXMLFile(outfile, target, 'name', entry.name)
    writeAVUToXMLFile(outfile, target, 'dc.source', entry.dcSource)
    writeAVUToXMLFile(outfile, target, 'dc.title', entry.dcTitle)
    writeAVUToXMLFile(outfile, target, 'dc.date', entry.dcDate.strftime(AssetProvenance.FMT_DATE))
    writeAVUToXMLFile(outfile, target, 'dc.publisher', entry.dcPublisher)
    writeAVUToXMLFile(outfile, target, 'dc.description', entry.dcDescription)
    
# Write point climate station metadata to the data file for that station
stations = GenericMetadata.readClimatePointStations(context)
for station in stations:
    target = collection + PATH_SEP_IRODS + station.data
    writeAVUToXMLFile(outfile, target, 'id', station.id)
    writeAVUToXMLFile(outfile, target, 'name', station.name)
    writeAVUToXMLFile(outfile, target, 'longitude', str(station.longitude), 'WGS84 degrees')
    writeAVUToXMLFile(outfile, target, 'latitude', str(station.latitude), 'WGS84 degrees')
    writeAVUToXMLFile(outfile, target, 'elevation', str(station.elevation), 'meters')

outfile.write('</metadata>\n')
outfile.close()


from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='ecohydrolib',
      version='1.20',
      description='Libraries and command-line scripts for performing ecohydrology data preparation workflows.',
      long_description=readme(),
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering :: GIS'        
      ],
      url='https://github.com/selimnairb/EcohydroLib',
      author='Brian Miles',
      author_email='brian_miles@unc.edu',
      license='BSD',
      packages=['ecohydrolib', 
                'ecohydrolib.climatedata',
                'ecohydrolib.geosciaus',
                'ecohydrolib.hydro1k',
                'ecohydrolib.nhdplus2',
                'ecohydrolib.nlcd',
                'ecohydrolib.solim',
                'ecohydrolib.spatialdata',
                'ecohydrolib.ssurgo',
                'ecohydrolib.tests',
                'ecohydrolib.wcs4dem' 
                ],
      install_requires=[
        #'GDAL', # GDAL does not build under OS X easy_install/PIP
        'pyproj',
        'numpy',
        'OWSLib',
        'oset',
        'httplib2',
        'shapely'
      ],
      scripts=['bin/DumpClimateStationInfo.py',
               'bin/DumpMetadataToiRODSXML.py',
               'bin/GenerateSoilPropertyRastersFromSOLIM.py',
               'bin/GenerateSoilPropertyRastersFromSSURGO.py',
               'bin/GetBoundingboxFromStudyareaShapefile.py',
               'bin/GetCatchmentShapefileForHYDRO1kBasins.py',
               'bin/GetCatchmentShapefileForNHDStreamflowGage.py',
               'bin/GetDEMExplorerDEMForBoundingbox.py',
               'bin/GetGADEMForBoundingBox.py',
               'bin/GetGHCNDailyClimateDataForBoundingboxCentroid.py',
               'bin/GetGHCNDailyClimateDataForStationsInBoundingbox.py',
               'bin/GetHYDRO1kDEMForBoundingbox.py',
               'bin/GetNHDStreamflowGageIdentifiersAndLocation.py',
               'bin/GetNLCDForDEMExtent.py',
               'bin/GetSSURGOFeaturesForBoundingbox.py',
               'bin/RegisterDEM.py',
               'bin/RegisterGage.py',
               'bin/RegisterRaster.py',
               'bin/RegisterStudyAreaShapefile.py',
               'bin/RunCmd.py',
               'bin/GHCNDSetup/GHCNDSetup.py',
               'bin/NHDPlusV2Setup/NHDPlusV2Setup.py',
               'cgi/GetCatchmentFeaturesForStreamflowGage',
               'cgi/LocateStreamflowGage'
      ],
      zip_safe=False)



#!/usr/bin/env python
"""@package RegisterStudyAreaShapefile

@brief Register shapefile into metadata store for a project directory,
copying the source shapefile into the project directory in the process (if the source
layer is not a shapefile, it will be converted on copy).  Study area will
be projected to WGS84 GCS (EPSG:4326)

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
None

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of metadata associated with the project directory:
   study_area_shapefile

2. Will write the following entry(ies) to the study area section of metadata associated with the project directory:
   bbox_wgs84

Usage:
@code
python ./RegisterStudyAreaShapefile.py -p /path/to/project_dir -s /shapefile/to/register
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import os
import sys
import errno
import argparse
import textwrap

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance
from ecohydrolib.spatialdata.utils import convertFeatureLayerToShapefile

# Handle command line options
parser = argparse.ArgumentParser(description='Register study area shapefile with project')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-s', '--studyAreaLayer', dest='studyAreaLayer', required=True,
                    help='The name of the study area feature layer to be registered.')
parser.add_argument('-b', '--publisher', dest='publisher', required=False,
                    help="The publisher of the DEM dataset, if not supplied 'SELF PUBLISHED' will be used")
parser.add_argument('--overwrite', dest='overwrite', action='store_true', required=False,
                    help='Overwrite existing datasets in the project.  If not specified, program will halt if a dataset already exists.')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if args.publisher:
    publisher = args.publisher
else:
    publisher = 'SELF PUBLISHED'

if not context.config.has_option('GDAL/OGR', 'PATH_OF_OGR2OGR'):
    sys.exit("Config file %s does not define option %s in section %s" & \
          (args.configfile, 'GDAL/OGR', 'PATH_OF_OGR2OGR'))

if not os.access(args.studyAreaLayer, os.R_OK):
    raise IOError(errno.EACCES, "Not allowed to read input study area %s" (args.demfile,))
inStudyAreaPath = os.path.abspath(args.studyAreaLayer)

shpFilename = convertFeatureLayerToShapefile(context.config,  context.projectDir, inStudyAreaPath, "studyarea", overwrite=args.overwrite)

# Write metadata
asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = 'study_area_shapefile'
asset.dcIdentifier = shpFilename
asset.dcSource = "file://%s" % (inStudyAreaPath,)
asset.dcTitle = 'Study area shapefile'
asset.dcPublisher = publisher
asset.dcDescription = cmdline
asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)



#!/usr/bin/env python
"""@package RunCmd

@brief Run an arbitrary command operating on data stored in a project directory,
with the command and its arguments captured in the metadata for the project.

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
None

Post conditions
---------------
1. Will write an entry to the history section of the project metadata

Usage:
@code
RunCmd.py -p /path/to/project_dir COMMAND [ARG1 ... ARGN]
@endcode
"""
import os
import sys
import errno
import argparse
import textwrap
import subprocess

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
import ecohydrolib.util

# Handle command line options
parser = argparse.ArgumentParser(description='Run arbitrary command against data in project, recording the command in history metadata')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                    help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('command')
parser.add_argument('args', nargs=argparse.REMAINDER)
args = parser.parse_args()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

# Run command
cmd = ecohydrolib.util.getAbsolutePathOfExecutable(args.command)
if cmd == None:
    sys.exit("Enable able to find command '%s'" % (args.command,) )

result = subprocess.call( [cmd] + args.args )
if result != 0:
    sys.exit("Command '%s' failed returning %d" % (args.command, result) )

# Build representation of command with absolute path of all command arguments
cmdline = cmd
for arg in args.args:
    cmdline += ' ' + ecohydrolib.util.getAbsolutePathOfItem(arg)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)




"""@package ecohydrolib.nhdplus2.webservice
    
@brief Methods for querying the NHDPlus V2 data set via custom web service

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os, errno
import sys

import socket
import httplib
import json
import textwrap
import tempfile, shutil

from ecohydrolib.spatialdata.utils import OGR_SHAPEFILE_DRIVER_NAME
from ecohydrolib.spatialdata.utils import OGR_DRIVERS

_DEFAULT_CRS = 'EPSG:4326'
_BUFF_LEN = 4096 * 10

HOST = 'ga-dev-wssi.renci.org'
URL_PROTO_GAGE_LOC = '/cgi-bin/LocateStreamflowGage?gageid={gageid}'
URL_PROTO_CATCHMENT = '/cgi-bin/GetCatchmentFeaturesForStreamflowGage?reachcode={reachcode}&measure={measure}'

CONTENT_TYPE = 'application/json'
CONTENT_TYPE_ERROR = 'text/plain'
RESPONSE_OK = 'OK'

class WebserviceError(Exception):
    def __init__(self, url, error):
        msg = "Encountered the following error when accessing URL %s, error: %s" % \
            (url, error)
        super(WebserviceError, self).__init__(msg)

def locateStreamflowGage(config, gageid):
    """ Query NHDPlus V2 web service for location information for streamflow gage
        listed in the NHDPlus V2 dataset
    
        @param config A Python ConfigParser (not currently used)
        @param gageid String representing streamflow gage ID (e.g. typically USGS site identifier)
        
        @return Tuple(A dictionary with keys: message, measure, reachcode, gage_lon, gage_lat 
        if location data were fetched or dictionary with key 'message' if not, URL of the request)
    """
    response = {}
    
    url = URL_PROTO_GAGE_LOC.format(gageid=gageid)
    urlFetched = "http://%s%s" % (HOST, url)
    
    conn = httplib.HTTPConnection(HOST)
    try:
        conn.request('GET', url)
        res = conn.getresponse(buffering=False)
    except socket.error as e:
        msg = "Encountered the following error when trying to get gage location from %s. Error: %s.  Please try again later or contact the developer." % \
            (urlFetched, str(e) )
        sys.stderr.write( textwrap.fill(msg) )
        return ( response, urlFetched )
    
    if 200 != res.status:
        msg = "HTTP response %d %s encountered when querying %s.  Please try again later or contact the developer." % \
            (res.status, res.reason, urlFetched)
        sys.stderr.write( textwrap.fill(msg) ) 
        return ( response, urlFetched )
    
    contentType = res.getheader('Content-Type')
    
    if contentType.find(CONTENT_TYPE) != -1:
        # The data returned were of the type expected, read the data
        response = json.load(res)
            
    else:
        msg = "Query from URL %s returned content type %s, was expecting type %s.  Operation failed." % \
            (urlFetched, contentType, CONTENT_TYPE)
        sys.stderr.write( textwrap.fill(msg) )
    
    return (response, urlFetched)

def getCatchmentFeaturesForStreamflowGage(config, outputDir,
                                          catchmentFilename, reachcode, measure,
                                          format=OGR_SHAPEFILE_DRIVER_NAME):
    """ Query NHDPlus V2 web service for features (in WGS 84) for 
        the drainage area associated with a given NHD (National 
        Hydrography Dataset) streamflow gage identified by a reach 
        code and measure.
        
        @param config A Python ConfigParser (not currently used)
        @param outputDir String representing the absolute/relative
        path of the directory into which output rasters should be
        written
        @param format String representing OGR driver to use
        @param catchmentFilename String representing name of file to
        save catchment features to.  The appropriate extension will be added to the file name
        @param reachcode String representing NHD streamflow gage 
        @param measure Float representing the measure along reach
        where Stream Gage is located in percent from downstream
        end of the one or more NHDFlowline features that are
        assigned to the ReachCode (see NHDPlusV21 GageLoc table)
        
        @return Tuple(String representing the name of the dataset in outputDir created to hold
        the features, URL of the request)
         
        @raise IOError(errno.EACCESS) if OGR binary is not executable
        @raise IOError(errno.ENOTDIR) if outputDir is not a directory
        @raise IOError(errno.EACCESS) if outputDir is not writable
        @raise Exception if output format is not known
        @raise WebserviceError if an error occurred calling the web service
    """
    ogrCmdPath = config.get('GDAL/OGR', 'PATH_OF_OGR2OGR')
    if not os.access(ogrCmdPath, os.X_OK):
        raise IOError(errno.EACCES, "The ogr2ogr binary at %s is not executable" %
                      ogrCmdPath)
    ogrCmdPath = os.path.abspath(ogrCmdPath)
    
    if not os.path.isdir(outputDir):
        raise IOError(errno.ENOTDIR, "Output directory %s is not a directory" % (outputDir,))
    if not os.access(outputDir, os.W_OK):
        raise IOError(errno.EACCES, "Not allowed to write to output directory %s" % (outputDir,))
    outputDir = os.path.abspath(outputDir)
    
    if not format in OGR_DRIVERS.keys():
        raise Exception("Output format '%s' is not known" % (format,) )
    
    catchmentFilename ="%s%s%s" % ( catchmentFilename, os.extsep, OGR_DRIVERS[format] )
    catchmentFilepath = os.path.join(outputDir, catchmentFilename)
    
    url = URL_PROTO_CATCHMENT.format( reachcode=reachcode, measure=str(measure) )
    urlFetched = "http://%s%s" % (HOST, url)
    
    conn = httplib.HTTPConnection(HOST)
    try:
        conn.request('GET', url)
        res = conn.getresponse(buffering=True)
    except socket.error as e:
        raise WebserviceError(urlFetched, str(e))
    
    if 200 != res.status:
        error = "%d %s" % (res.status, res.reason)
        raise WebserviceError(urlFetched, error)
    
    contentType = res.getheader('Content-Type')
    
    if contentType.find(CONTENT_TYPE_ERROR) != -1:
        error = res.read()
        raise WebserviceError(urlFetched, error)
    
    elif contentType.find(CONTENT_TYPE) != -1:
        # The data returned were of the type expected, read the data
        tmpdir = tempfile.mkdtemp()
        
        failure = False
        data = res.read(_BUFF_LEN)
        if data:            
            tmpfile = os.path.join( tmpdir, 'catchment.geojson' )
            f = open(tmpfile, 'wb')
            while data:
                f.write(data)
                data = res.read(_BUFF_LEN)
            f.close()
            # Convert GeoJSON to ESRI Shapfile using OGR
            ogrCommand = "%s -s_srs EPSG:4326 -t_srs EPSG:4326 -f '%s' %s %s" % (ogrCmdPath, format, catchmentFilepath, tmpfile)
            os.system(ogrCommand)
            if not os.path.exists(catchmentFilepath):
                failure = False
        
        shutil.rmtree(tmpdir)
        if failure:
            raise WebserviceError(urlFetched, "Failed to store catchment features in file %s" % (catchmentFilepath,) )
    else:
        error = "Recieved content type %s but expected type %s" % (contentType, CONTENT_TYPE)
        raise WebserviceError(urlFetched, error)
    
    return (catchmentFilename, urlFetched)


"""@package ecohydrolib.grasslib
    
@brief Class that encapsulated environment setup for using GRASS GIS scripting and low-level APIs

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
"""
import os, errno
import sys
import shutil
import tempfile
import importlib

GRASS_RASTER_RESAMPLE_NEAREST = 'nearest'
GRASS_RASTER_RESAMPLE_BILINEAR = 'bilinear'
GRASS_RASTER_RESAMPLE_BICUBIC = 'bicubic'
GRASS_RASTER_RESAMPLE_METHODS = [GRASS_RASTER_RESAMPLE_NEAREST, 
                                 GRASS_RASTER_RESAMPLE_BILINEAR,
                                 GRASS_RASTER_RESAMPLE_BICUBIC]

DEFAULT_LOCATION = 'default'
DEFAULT_MAPSET = 'PERMANENT'

class GRASSConfig(object):
    def __init__(self, context, dbase, location=None, mapset=None, newLocation=False, overwrite=False):
        """ @brief Constructor for GRASSConfig
        
            @param context ecohydrolib.Context with config containing a GRASS section with GISBASE entry
            @param dbase String representing a GRASS GIS data directory
            @param location String representing a GRASS project location within the data directory; defaults to DEFAULT_LOCATION.
            @param mapset String representing a GRASS mapset within the location; defaults to DEFAULT_MAPSET.
            @param newLocation Boolean, True if this is a new location, False if the location is expected to already exist
            @param overwrite Boolean, if False, constructor will raise an error if newLocation is True and the location already exists
            
            @raise IOError(errno.ENOTDIR) if dbase is not a writable directory
            @raise IOError(errno.EACCESS) if dbase is not writable
            @raise IOError(errno.EEXIST) if location exists and newLocation == True
        """
        self.gisbase = context.config.get('GRASS', 'GISBASE')
        if not location:
            location = DEFAULT_LOCATION
        if not mapset:
            mapset = DEFAULT_MAPSET
        # Make sure dbase exists
        self.dbase = os.path.abspath(dbase)
        if not os.path.exists(self.dbase):
            (grassDbLoc, grassDbName) = os.path.split(self.dbase)
            if not (os.path.isdir(grassDbLoc) and os.access(grassDbLoc, os.W_OK)):
                raise IOError(errno.ENOTDIR, "%s is not a writable directory" % (grassDbLoc,))
            os.makedirs(self.dbase)
        else:
            if not os.access(self.dbase, os.W_OK):
                raise(errno.EACCES, "Not allowed to write to %s" % (self.dbase,))
        # Check if location already exists, if it does, raise an error if newLocation == True
        self.location = location
    
        locationPath = os.path.join(self.dbase, self.location)
        if os.path.exists( locationPath ):
            if newLocation:
                if overwrite:
                    # Delete existing location
                    shutil.rmtree(locationPath)
                else:
                    raise IOError(errno.EEXIST, "Location '%s' already exists in %s" % \
                                  (location, self.dbase))
        self.mapset = mapset

class GRASSLib(object): 
    def __init__(self, grassConfig=None, grassScripting=None, grassAPI=None):
        """ @brief Constructor for GRASSLib
        
            @param grassConfig GRASSConfig instance 
            @param grassScripting Previously imported grass.script (GRASS scripting API), 
            if None, grass.script will be imported
            @param api Previously imported grass.lib.gis (low-level GRASS API); if None
            grass.lib.gis will be imported
        """
        self.grassConfig = grassConfig
        
        if not grassScripting:
            self.script = self._setupGrassScriptingEnvironment()
        else:
            self.script = grassScripting
            
        if not grassAPI:
            self.api = self._setupGrassEnvironment()
        else:
            self.api = grassAPI 

    def _setupGrassScriptingEnvironment(self):
        """ @brief Set up GRASS environment for using GRASS scripting API from 
            Python (e.g. grass.script)
        """
        os.environ['GISBASE'] = self.grassConfig.gisbase
        sys.path.append(os.path.join(self.grassConfig.gisbase, 'etc', 'python'))
        import grass.script.setup as gsetup
        gsetup.init(self.grassConfig.gisbase, \
                    self.grassConfig.dbase, self.grassConfig.location, \
                    self.grassConfig.mapset)

        self.script = importlib.import_module('grass.script')
        return self.script
    
    def _setupGrassEnvironment(self):
        """ @brief Set up GRASS environment for using GRASS low-level API from 
            Python (e.g. grass.lib)
        """
        gisBase = self.grassConfig.gisbase
        sys.path.append(os.path.join(gisBase, 'etc', 'python'))
        os.environ['LD_LIBRARY_PATH'] = os.path.join(gisBase, 'lib')
        os.environ['DYLD_LIBRARY_PATH'] = os.path.join(gisBase, 'lib')
        # Write grassrc
        os.environ['GISRC'] = self._initializeGrassrc()
        os.environ['GIS_LOCK'] = str(os.getpid())
        self.api = importlib.import_module('grass.lib.gis')
        self.api.G_gisinit('')
        return self.api
    
    def _initializeGrassrc(self):
        grassRcFile = tempfile.NamedTemporaryFile(prefix='grassrc-', delete=False)
        grassRcContent = "GISDBASE: %s\nLOCATION_NAME: %s\nMAPSET: %s\n" % \
            (self.grassConfig.dbase, self.grassConfig.location, self.grassConfig.mapset)
        grassRcFile.write(grassRcContent)
        return grassRcFile.name



#!/usr/bin/env python
"""@package GetNHDStreamflowGageIdentifiersAndLocation
 
@brief Get NHDPlus2 streamflow gage identifiers (reachcode, measure along reach in percent) for a USGS gage 
@brief Get lat/lon, in WGS84 (EPSG:4326), from gage point layer (Gage_Loc) for gage identified by a source_fea (e.g. USGS Site Number) 

This software is provided free of charge under the New BSD License. Please see
the following license information:

Copyright (c) 2013, University of North Carolina at Chapel Hill
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the University of North Carolina at Chapel Hill nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF NORTH CAROLINA AT CHAPEL HILL
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


@author Brian Miles <brian_miles@unc.edu>
  
  
Pre conditions
--------------
1. Configuration file must define the following sections and values:
   'NHDPLUS2', 'PATH_OF_NHDPLUS2_DB'
   'NHDPLUS2', 'PATH_OF_NHDPLUS2_GAGELOC'

Post conditions
---------------
1. Will write the following entry(ies) to the manifest section of
metadata associated with the project directory:
   gage [the name of the streamflow gage shapefile]  
   
2. Will write the following entry(ies) to the study area section of
metadata associated with the project directory:
   gage_id_attr [the name of the attribute in the gage shapefile that
     uniquely identifies a streamflow gage]
   gage_id
   nhd_gage_reachcode
   nhd_gage_measure_pct
   gage_lat_wgs84
   gage_lon_wgs84

Usage:
@code
GetNHDStreamflowGageIdentifiersAndLocation.py -p /path/to/project_dir -g 01589330
@endcode

@note EcohydroLib configuration file must be specified by environmental variable 'ECOHYDROWORKFLOW_CFG',
or -i option must be specified. 
"""
import sys
import argparse

from ecohydrolib.context import Context
from ecohydrolib.metadata import GenericMetadata
from ecohydrolib.metadata import AssetProvenance

from ecohydrolib.nhdplus2.webservice import locateStreamflowGage
from ecohydrolib.nhdplus2.webservice import RESPONSE_OK
from ecohydrolib.nhdplus2.networkanalysis import getNHDReachcodeAndMeasureForGageSourceFea
from ecohydrolib.nhdplus2.networkanalysis import getLocationForStreamGageByGageSourceFea
from ecohydrolib.spatialdata.utils import writeCoordinatePairsToPointShapefile

# Handle command line options
parser = argparse.ArgumentParser(description='Get NHDPlus2 streamflow gage identifiers for a USGS gage.')
parser.add_argument('-i', '--configfile', dest='configfile', required=False,
                    help='The configuration file')
parser.add_argument('-p', '--projectDir', dest='projectDir', required=True,
                  help='The directory to which metadata, intermediate, and final files should be saved')
parser.add_argument('-s', '--source', dest='source', required=False, choices=['local', 'webservice'], default='webservice',
                    help='Source to query NHDPlusV2 dataset')
parser.add_argument('-g', '--gageid', dest='gageid', required=True,
                    help='An integer representing the USGS site identifier')
args = parser.parse_args()
cmdline = GenericMetadata.getCommandLine()

configFile = None
if args.configfile:
    configFile = args.configfile

context = Context(args.projectDir, configFile) 

if args.source == 'local':
    sys.stdout.write('Getting identifiers and location from local NHDPlus dataset...')
    sys.stdout.flush()
    if not context.config.has_option('NHDPLUS2', 'PATH_OF_NHDPLUS2_DB'):
        sys.exit("Config file %s does not define option %s in section %s" & \
              (args.configfile, 'NHDPLUS2', 'PATH_OF_NHDPLUS2_DB'))
    
    if not context.config.has_option('NHDPLUS2', 'PATH_OF_NHDPLUS2_GAGELOC'):
        sys.exit("Config file %s does not define option %s in section %s" & \
              (args.configfile, 'NHDPLUS2', 'PATH_OF_NHDPLUS2_GAGELOC'))
    
    result = getNHDReachcodeAndMeasureForGageSourceFea(context.config, args.gageid)
    if result:
        reachcode = result[0]
        measure = result[1]
    else:
        sys.exit("Gage '%s' not found" % (args.gageid,) )
    
    result = getLocationForStreamGageByGageSourceFea(context.config, args.gageid)
    if result:
        gage_lat = result[1]
        gage_lon = result[0]
    else:
        sys.exit("Gage '%s' not found" % (args.gageid,) )
        
    source = 'http://www.horizon-systems.com/NHDPlus/NHDPlusV2_home.php'
    sys.stdout.write('done\n')
else:
    # Query NHDPlusV2 dataset via web service
    sys.stdout.write('Geting identifiers and location via NHDPlus webservice...')
    sys.stdout.flush()
    (response, source) = locateStreamflowGage(context.config, args.gageid)
    if response['message'] != RESPONSE_OK:
        sys.exit("Failed to get gage identifiers/location for gage '%s', webservice response: %s" % \
                 (args.gageid, response['message']) )
    else:
        reachcode = response['reachcode']
        measure = response['measure']
        gage_lat = response['gage_lat']
        gage_lon = response['gage_lon']
        
    sys.stdout.write('done\n')    

# Write gage coordinates to a shapefile in the project directory
shpFilename = writeCoordinatePairsToPointShapefile(context.projectDir, "gage", 
                                                   "gage_id", [args.gageid], [(gage_lon, gage_lat)])

# Write study area metadata
GenericMetadata.writeStudyAreaEntry(context, 'gage_id_attr', 'gage_id')
GenericMetadata.writeStudyAreaEntry(context, 'gage_id', args.gageid)
GenericMetadata.writeStudyAreaEntry(context, 'nhd_gage_reachcode', reachcode)
GenericMetadata.writeStudyAreaEntry(context, 'nhd_gage_measure_pct', measure)
GenericMetadata.writeStudyAreaEntry(context, 'gage_lat_wgs84', gage_lat)
GenericMetadata.writeStudyAreaEntry(context, 'gage_lon_wgs84', gage_lon)

# Write provenance
asset = AssetProvenance(GenericMetadata.MANIFEST_SECTION)
asset.name = 'gage'
asset.dcIdentifier = shpFilename
asset.dcSource = source
asset.dcTitle = 'Streamflow gage'
asset.dcPublisher = 'USGS'
asset.dcDescription = cmdline
asset.writeToMetadata(context)

# Write processing history
GenericMetadata.appendProcessingHistoryItem(context, cmdline)
